import 'dart:typed_data';
import 'package:vid/us/vid_us_application.dart';
import 'package:vid/us/vid_us_data_http_reader.dart';
import 'package:vid/us/vid_us_data_reader.dart';
import 'package:vid/us/vid_us_data_writer.dart';
import 'package:vid/us/vid_us_image.dart';
import 'package:vid/us/vid_us_probe.dart';

enum VidUsImageFormat {
  Jpeg,
  Png,
  H264,
  Zip,
  Diff,
}

///Not suport carotid3d for now.
class VidUsImageData {
  late VidUsDataReader _reader;
  late List<int> _imagePositionList;
  late int _version;
  late int _imageCount;
  late VidUsProbe _probe;
  late VidUsImageFormat _imageFormat;
  late Uint8List _extendedData;
  late List<VidUsImage> _imagesList;

  final String header = "VINNO IMAGE DATA";

  /// Gets the version of this image data.
  int get version => _version;

  /// Gets the image count of this image data.
  int get imageCount => _imageCount;

  /// Gets the probe information.
  VidUsProbe get probe => _probe;

  /// Gets the image format of this image data.
  VidUsImageFormat get imageFormat => _imageFormat;

  /// Gets or sets the extended data.
  Uint8List get extendedData => _extendedData;

  /// Create a VINNO Image Data.
  VidUsImageData(Uint8List data) {
    _reader = VidUsDataReader(data);
    var header = _reader.readString();
    if (header != header) {
      throw Exception("The input data is not a VID data.");
    }
    _version = _reader.readInt();
    //Get probe info
    var probeData = _reader.readBytes();
    _probe = VidUsProbe.fromBytes(probeData);
    _imageFormat = VidUsImageFormat.values[_reader.readInt()];
    _extendedData = _reader.readBytes();

    _imagePositionList = [];
    var imagePositionListData = _reader.readBytes();
    _imageCount = imagePositionListData.length ~/ 8;
    var imagePositionReader = VidUsDataReader(imagePositionListData);
    for (var i = 0; i < _imageCount; i++) {
      _imagePositionList.add(imagePositionReader.readInt64V2());
    }
    _imagesList = [];
  }

  /// Create an empty third part VINNO Image Data
  VidUsImageData.empty(int fps, VidUsImageFormat imageFormat) {
    _imageCount = 0;
    _version = 1;
    _probe = VidUsProbe(
      "ThirdPart",
      VidUsProbeType.Linear,
      VidUsApplication(
        "ThirdPart",
        "ThirdPart",
        "ThirdPart",
        "ThirdPart",
        false,
      ),
      fps.toDouble(),
    );
    _imageFormat = imageFormat;
    _extendedData = Uint8List(10);
    _imagePositionList = [];
    _imagesList = [];
  }

  /// Add an image to the image data.
  void addJpegImage(Uint8List jpegBytes, int width, int height) {
    if (_imageFormat != VidUsImageFormat.Jpeg) {
      throw Exception("The image format is not Jpeg.");
    }
    if (_imagesList.length != _imageCount) {
      _fillImageList();
    }
    VidUsImage image = VidUsImage(_imageCount, width, height, jpegBytes);
    _imageCount++;
    _imagesList.add(image);
  }

  Uint8List flush() {
    if (_imageCount != _imagesList.length) {
      _fillImageList();
    }
    var imagePositionOffset = 0;
    var writer = VidUsDataWriter();
    writer.writeString(header);
    writer.writeInt(_version);
    writer.writeBytes(_probe.toBytes());
    writer.writeInt(_imageFormat.index);
    writer.writeBytes(_extendedData);
    imagePositionOffset = writer.data.length + _imageCount * 8 + 4;
    var imagePositionWriter = VidUsDataWriter();
    for (var i = 0; i < _imageCount; i++) {
      imagePositionWriter.writeInt64V2(imagePositionOffset);
      imagePositionOffset += _imagesList[i].toBytes().length + 4;
    }
    writer.writeBytes(imagePositionWriter.data);
    for (var i = 0; i < _imageCount; i++) {
      writer.writeBytes(_imagesList[i].toBytes());
    }
    return writer.data;
  }

  /// Get one image from the vid.
  VidUsImage getImage(int index) {
    if (index >= _imageCount || index < 0) {
      throw Exception("Can not find image Data");
    }
    if (index < _imagesList.length) {
      return _imagesList[index];
    }
    //Jump to image.
    var imageData = _reader.readBytes(_imagePositionList[index]);
    return VidUsImage.fromBytes(imageData);
  }

  /// Fill all images to the image list.
  void _fillImageList() {
    _imagesList = [];
    for (var i = 0; i < _imageCount; i++) {
      _imagesList.add(getImage(i));
    }
  }

  void setThirdPartFPS(int newFps) {
    _probe = VidUsProbe(
      "ThirdPart",
      VidUsProbeType.Linear,
      VidUsApplication(
        "ThirdPart",
        "ThirdPart",
        "ThirdPart",
        "ThirdPart",
        false,
      ),
      newFps.toDouble(),
    );
  }
}

///Raised when getting data from the downloaded data timeout. depends on the readHeaderTimeout and
///readImageTimeout parameters.
class ReadTimeoutException implements Exception {}

///Raised when the Http operation already closed.
class AlreadyClosedException implements Exception {}

///Only suport read for now.
///Not suport carotid3d for now.
class HttpVidUsImageData {
  late VidUsDataHttpReader _reader;
  late List<int> _imagePositionList;
  late int _version;
  late int _imageCount;
  late VidUsProbe _probe;
  late VidUsImageFormat _imageFormat;
  late Uint8List _extendedData;
  late int _readHeaderTimeout;
  late int _readImageTimeout;
  late bool _initialized;
  late bool _closed;

  final Duration delayDuration = const Duration(milliseconds: 10);
  final String header = "VINNO IMAGE DATA";

  /// Gets the version of this image data.
  int get version => _version;

  /// Gets the image count of this image data.
  int get imageCount => _imageCount;

  /// Gets the probe information.
  VidUsProbe get probe => _probe;

  /// Gets the image format of this image data.
  VidUsImageFormat get imageFormat => _imageFormat;

  /// Gets the extended data.
  Uint8List get extendedData => _extendedData;

  /// Gets the initialized state.
  bool get initialized => _initialized;

  /// Create a VINNO Image Data.
  /// [readHeaderTimeout] is the timeout value of reading header and probe information etc..., we need give as much time as possible
  /// to let the reader read the header.
  /// [readImageTimeout] is the timeout value of reading one frame.
  /// [minChunkSize] The min download chunk size, set to a large value may speedup the download speed, but will cause the progress not smooth.
  HttpVidUsImageData(String url,
      {DownloadCallback? downloadCallback,
      int minChunkSize = 65536,
      int readHeaderTimeout = 3000,
      int readImageTimeout = 500}) {
    _initialized = false;
    _closed = false;
    _readHeaderTimeout = readHeaderTimeout;
    _readImageTimeout = readImageTimeout;
    _reader = VidUsDataHttpReader(url,
        downloadCallback: downloadCallback, minChunkSize: minChunkSize);
  }

  ///Get the downloaded data of this HttpVidUsImageData
  Uint8List getDownloadedData() {
    return _reader.toBytes();
  }

  ///Initialize the HttpVidUsImageData, must be called firstly.
  Future initialize() async {
    if (!_initialized) {
      var header = await _readHeader();
      if (header != header) {
        throw Exception("The input data is not a VID data.");
      }
      _version = await _readVersion();
      //Get probe info
      _probe = await _readProbe();
      _imageFormat = await _readImageFormat();
      _extendedData = await _readExtendedData();
      _imagePositionList = [];
      var imagePositionListData = await _readImagePositionListData();
      _imageCount = imagePositionListData.length ~/ 8;
      var imagePositionReader = VidUsDataReader(imagePositionListData);
      for (var i = 0; i < _imageCount; i++) {
        _imagePositionList.add(imagePositionReader.readInt64V2());
      }
      _initialized = true;
    }
  }

  ///Close the HttpVidUsImageData, it will force close the download operation and reading operation.
  void close() {
    _closed = true;
    _reader.close();
  }

  ///Read header from http data.
  Future<String> _readHeader() async {
    var timeout = 0;
    while (!_closed) {
      try {
        var header = _reader.readString();
        return header;
      } catch (ex) {
        if (ex is NotReadyException) {
          if (timeout >= _readHeaderTimeout) {
            throw ReadTimeoutException();
          }
          await Future.delayed(delayDuration);
          timeout += delayDuration.inMilliseconds;
        } else {
          rethrow;
        }
      }
    }
    throw AlreadyClosedException();
  }

  ///Read the version from http data.
  Future<int> _readVersion() async {
    var timeout = 0;
    while (!_closed) {
      try {
        var version = _reader.readInt();
        return version;
      } catch (ex) {
        if (ex is NotReadyException) {
          if (timeout >= _readHeaderTimeout) {
            throw ReadTimeoutException();
          }
          await Future.delayed(delayDuration);
          timeout += delayDuration.inMilliseconds;
        } else {
          rethrow;
        }
      }
    }
    throw AlreadyClosedException();
  }

  ///Read the probe info from http data.
  Future<VidUsProbe> _readProbe() async {
    var timeout = 0;
    while (!_closed) {
      try {
        var probeData = _reader.readBytes();
        return VidUsProbe.fromBytes(probeData);
      } catch (ex) {
        if (ex is NotReadyException) {
          if (timeout >= _readHeaderTimeout) {
            throw ReadTimeoutException();
          }
          await Future.delayed(delayDuration);
          timeout += delayDuration.inMilliseconds;
        } else {
          rethrow;
        }
      }
    }
    throw AlreadyClosedException();
  }

  ///Read the image format from http data.
  Future<VidUsImageFormat> _readImageFormat() async {
    var timeout = 0;
    while (!_closed) {
      try {
        return VidUsImageFormat.values[_reader.readInt()];
      } catch (ex) {
        if (ex is NotReadyException) {
          if (timeout >= _readHeaderTimeout) {
            throw ReadTimeoutException();
          }
          await Future.delayed(delayDuration);
          timeout += delayDuration.inMilliseconds;
        } else {
          rethrow;
        }
      }
    }
    throw AlreadyClosedException();
  }

  ///Read extended data from http data.
  Future<Uint8List> _readExtendedData() async {
    var timeout = 0;
    while (!_closed) {
      try {
        return _reader.readBytes();
      } catch (ex) {
        if (ex is NotReadyException) {
          if (timeout >= _readHeaderTimeout) {
            throw ReadTimeoutException();
          }
          await Future.delayed(delayDuration);
          timeout += delayDuration.inMilliseconds;
        } else {
          rethrow;
        }
      }
    }
    throw AlreadyClosedException();
  }

  ///Read the image positions data from http data.
  Future<Uint8List> _readImagePositionListData() async {
    var timeout = 0;
    while (!_closed) {
      try {
        return _reader.readBytes();
      } catch (ex) {
        if (ex is NotReadyException) {
          if (timeout >= _readHeaderTimeout) {
            throw ReadTimeoutException();
          }
          await Future.delayed(delayDuration);
          timeout += delayDuration.inMilliseconds;
        } else {
          rethrow;
        }
      }
    }
    throw AlreadyClosedException();
  }

  /// Get one image from the vid.
  Future<VidUsImage> getImage(int index) async {
    if (index >= _imageCount || index < 0) {
      throw Exception("Can not find image Data");
    }
    //Jump to image.
    var timeout = 0;
    while (!_closed) {
      try {
        var imageData = _reader.readBytes(_imagePositionList[index]);
        return VidUsImage.fromBytes(imageData);
      } catch (ex) {
        if (ex is NotReadyException) {
          if (timeout >= _readImageTimeout) {
            throw ReadTimeoutException();
          }
          await Future.delayed(delayDuration);
          timeout += delayDuration.inMilliseconds;
        } else {
          rethrow;
        }
      }
    }
    throw AlreadyClosedException();
  }
}
